# Value of dataset.status if dataset is published
DS_PUBLISHED_STATUS = 'published'

# Name of required source table arg we auto populate
SOURCE_TABLE_ARG_NAME = 'source_table'

# Transform Arg Types which expect DS inputs
# should contain Datasets Objs
TABLE_ARG_TYPE = 'table'
TABLE_LIST_ARG_TYPE = 'table_list'
